@@ -258,9 +258,11 @@ void CSoundFile::UpgradeModule()
 
 	if(m_dwLastSavedWithVersion < MPT_V("1.20.00.00"))
 	{
-		for(INSTRUMENTINDEX i = 1; i <= GetNumInstruments(); i++) if(Instruments[i] != nullptr)
+		for(INSTRUMENTINDEX i = 1; i <= GetNumInstruments(); i++)
 		{
 			ModInstrument *ins = Instruments[i];
+			if(!ins)
+				continue;
 			// Previously, volume swing values ranged from 0 to 64. They should reach from 0 to 100 instead.
 			ins->nVolSwing = static_cast<uint8>(std::min(static_cast<uint32>(ins->nVolSwing * 100 / 64), uint32(100)));
 
@@ -268,7 +270,7 @@ void CSoundFile::UpgradeModule()
 			{
 				// Previously, Pitch/Pan Separation was only half depth (plot twist: it was actually only quarter depth).
 				// This was corrected in compatible mode in OpenMPT 1.18, and in OpenMPT 1.20 it is corrected in normal mode as well.
-				ins->nPPS = (ins->nPPS + (ins->nPPS >= 0 ? 1 : -1)) / 2;
+				ins->nPPS = static_cast<int8>((ins->nPPS + (ins->nPPS >= 0 ? 1 : -1)) / 2);
 			}
 
 			if(!compatModeIT || m_dwLastSavedWithVersion < MPT_V("1.17.03.02"))
@@ -350,8 +352,21 @@ void CSoundFile::UpgradeModule()
 		}
 	}
 
+	bool hasAnyPlugins = false;
+	if(GetType() & (MOD_TYPE_IT | MOD_TYPE_MPT | MOD_TYPE_XM))
+	{
+		for(auto &plugin : m_MixPlugins)
+		{
+			if(plugin.IsValidPlugin())
+			{
+				hasAnyPlugins = true;
+				break;
+			}
+		}
+	}
+
 #ifndef NO_PLUGINS
-	if(m_dwLastSavedWithVersion < MPT_V("1.22.07.01"))
+	if(m_dwLastSavedWithVersion < MPT_V("1.22.07.01") && hasAnyPlugins)
 	{
 		// Convert ANSI plugin path names to UTF-8 (irrelevant in probably 99% of all cases anyway, I think I've never seen a VST plugin with a non-ASCII file name)
 		for(auto &plugin : m_MixPlugins)
@@ -405,26 +420,30 @@ void CSoundFile::UpgradeModule()
 
 	if(m_dwLastSavedWithVersion < MPT_V("1.26.00.00"))
 	{
-		for(INSTRUMENTINDEX i = 1; i <= GetNumInstruments(); i++) if(Instruments[i] != nullptr)
+		for(INSTRUMENTINDEX i = 1; i <= GetNumInstruments(); i++)
 		{
 			ModInstrument *ins = Instruments[i];
+			if(!ins)
+				continue;
 			// Even after fixing it in OpenMPT 1.18, instrument PPS was only half the depth.
-			ins->nPPS = (ins->nPPS + (ins->nPPS >= 0 ? 1 : -1)) / 2;
+			ins->nPPS = static_cast<int8>((ins->nPPS + (ins->nPPS >= 0 ? 1 : -1)) / 2);
 
 			// OpenMPT 1.18 fixed the depth of random pan in compatible mode.
 			// OpenMPT 1.26 fixes it in normal mode too.
 			if(!compatModeIT || m_dwLastSavedWithVersion < MPT_V("1.18.00.00"))
-			{
-				ins->nPanSwing = (ins->nPanSwing + 3) / 4u;
-			}
+				ins->nPanSwing = static_cast<uint8>((ins->nPanSwing + 3) / 4u);
+
+			// Before OpenMPT 1.26 (r6129), it was possible to trigger MIDI notes using channel plugins if the instrument had a valid MIDI channel.
+			if(!ins->nMixPlug && ins->HasValidMIDIChannel() && m_dwLastSavedWithVersion >= MPT_V("1.17.00.00"))
+				m_playBehaviour.set(kMIDINotesFromChannelPlugin);
 		}
 	}
 
 	if(m_dwLastSavedWithVersion < MPT_V("1.28.00.12"))
 	{
-		for(INSTRUMENTINDEX i = 1; i <= GetNumInstruments(); i++) if(Instruments[i] != nullptr)
+		for(INSTRUMENTINDEX i = 1; i <= GetNumInstruments(); i++)
 		{
-			if(Instruments[i]->VolEnv.nReleaseNode != ENV_RELEASE_NODE_UNSET)
+			if(Instruments[i] != nullptr && Instruments[i]->VolEnv.nReleaseNode != ENV_RELEASE_NODE_UNSET)
 			{
 				m_playBehaviour.set(kLegacyReleaseNode);
 				break;
@@ -434,9 +453,9 @@ void CSoundFile::UpgradeModule()
 
 	if(m_dwLastSavedWithVersion < MPT_V("1.28.03.04"))
 	{
-		for(INSTRUMENTINDEX i = 1; i <= GetNumInstruments(); i++) if (Instruments[i] != nullptr)
+		for(INSTRUMENTINDEX i = 1; i <= GetNumInstruments(); i++)
 		{
-			if(Instruments[i]->pluginVolumeHandling == PLUGIN_VOLUMEHANDLING_MIDI || Instruments[i]->pluginVolumeHandling == PLUGIN_VOLUMEHANDLING_DRYWET)
+			if(Instruments[i] != nullptr && (Instruments[i]->pluginVolumeHandling == PLUGIN_VOLUMEHANDLING_MIDI || Instruments[i]->pluginVolumeHandling == PLUGIN_VOLUMEHANDLING_DRYWET))
 			{
 				m_playBehaviour.set(kMIDIVolumeOnNoteOffBug);
 				break;
@@ -578,6 +597,14 @@ void CSoundFile::UpgradeModule()
 			{ kITDCTBehaviour,                MPT_V("1.29.00.57") },
 			{ kITPitchPanSeparation,          MPT_V("1.30.00.53") },
 			{ kITResetFilterOnPortaSmpChange, MPT_V("1.30.08.02") },
+			{ kITInitialNoteMemory,           MPT_V("1.31.00.25") },
+			{ kITNoSustainOnPortamento,       MPT_V("1.32.00.13") },
+			{ kITEmptyNoteMapSlotIgnoreCell,  MPT_V("1.32.00.13") },
+			{ kITOffsetWithInstrNumber,       MPT_V("1.32.00.15") },
+			{ kITDoublePortamentoSlides,      MPT_V("1.32.00.27") },
+			{ kITCarryAfterNoteOff,           MPT_V("1.32.00.40") },
+			{ kITNoteCutWithPorta,            MPT_V("1.32.01.02") },
+			{ kITVolColNoSlidePropagation,    MPT_V("1.32.02.03") },
 		};
 
 		for(const auto &b : behaviours)
@@ -601,6 +628,8 @@ void CSoundFile::UpgradeModule()
 			{ kFT2NoteDelayWithoutInstr,     MPT_V("1.28.00.44") },
 			{ kITFT2DontResetNoteOffOnPorta, MPT_V("1.29.00.34") },
 			{ kFT2PortaResetDirection,       MPT_V("1.30.00.40") },
+			{ kFT2AutoVibratoAbortSweep,     MPT_V("1.32.00.29") },
+			{ kFT2OffsetMemoryRequiresNote,  MPT_V("1.32.00.43") },
 		};
 
 		for(const auto &b : behaviours)
@@ -624,6 +653,7 @@ void CSoundFile::UpgradeModule()
 			{ kST3RetrigAfterNoteCut,      MPT_V("1.29.00.00") },
 			{ kFT2ST3OffsetOutOfRange,     MPT_V("1.29.00.00") },
 			{ kApplyUpperPeriodLimit,      MPT_V("1.30.00.45") },
+			{ kST3TonePortaWithAdlibNote,  MPT_V("1.31.00.13") },
 		};
 
 		for(const auto &b : behaviours)
@@ -686,9 +716,10 @@ void CSoundFile::UpgradeModule()
 		&& m_dwLastSavedWithVersion >= MPT_V("1.23.01.02") && m_dwLastSavedWithVersion < MPT_V("1.28.00.43"))
 	{
 		// Bug that effectively clamped the release node to the sustain end
-		for(INSTRUMENTINDEX i = 1; i <= GetNumInstruments(); i++) if(Instruments[i] != nullptr)
+		for(INSTRUMENTINDEX i = 1; i <= GetNumInstruments(); i++)
 		{
-			if(Instruments[i]->VolEnv.nReleaseNode != ENV_RELEASE_NODE_UNSET
+			if(Instruments[i] != nullptr
+				&& Instruments[i]->VolEnv.nReleaseNode != ENV_RELEASE_NODE_UNSET
 				&& Instruments[i]->VolEnv.dwFlags[ENV_SUSTAIN]
 				&& Instruments[i]->VolEnv.nReleaseNode > Instruments[i]->VolEnv.nSustainEnd)
 			{
@@ -717,7 +748,8 @@ void CSoundFile::UpgradeModule()
 		}
 	}
 
-	if(m_dwLastSavedWithVersion >= MPT_V("1.27.00.42") && m_dwLastSavedWithVersion < MPT_V("1.30.00.46") && (GetType() & (MOD_TYPE_IT | MOD_TYPE_MPT | MOD_TYPE_XM)))
+#ifndef NO_PLUGINS
+	if(m_dwLastSavedWithVersion >= MPT_V("1.27.00.42") && m_dwLastSavedWithVersion < MPT_V("1.30.00.46") && hasAnyPlugins)
 	{
 		// The Flanger DMO plugin is almost identical to the Chorus... but only almost.
 		// The effect implementation was the same in OpenMPT 1.27-1.29, now it isn't anymore.
@@ -729,7 +761,26 @@ void CSoundFile::UpgradeModule()
 		}
 	}
 
-	if(m_dwLastSavedWithVersion >= MPT_V("1.27") && m_dwLastSavedWithVersion < MPT_V("1.30.06.00") && (GetType() & (MOD_TYPE_IT | MOD_TYPE_MPT | MOD_TYPE_XM)))
+	if(m_dwLastSavedWithVersion < MPT_V("1.30.00.54") && hasAnyPlugins)
+	{
+		// Currently active program and bank is assumed to be 1 when starting playback
+		for(INSTRUMENTINDEX i = 1; i <= GetNumInstruments(); i++)
+		{
+			if(Instruments[i] && (Instruments[i]->nMidiProgram == 1 || Instruments[i]->wMidiBank == 1))
+			{
+				m_playBehaviour.set(kPluginDefaultProgramAndBank1);
+				break;
+			}
+		}
+	}
+
+	if(m_dwLastSavedWithVersion < MPT_V("1.31.00.09") && hasAnyPlugins)
+	{
+		// Old-style plugin tone portamento
+		m_playBehaviour.set(kPluginIgnoreTonePortamento);
+	}
+
+	if(m_dwLastSavedWithVersion >= MPT_V("1.27") && m_dwLastSavedWithVersion < MPT_V("1.30.06.00") && hasAnyPlugins)
 	{
 		// Fix off-by-one delay length in older Echo DMO emulation
 		for(auto &plugin : m_MixPlugins)
@@ -746,6 +797,31 @@ void CSoundFile::UpgradeModule()
 			}
 		}
 	}
+
+	if(m_dwLastSavedWithVersion >= MPT_V("1.17") && m_dwLastSavedWithVersion < MPT_V("1.32.00.30") && hasAnyPlugins)
+	{
+		for(const auto &plugin : m_MixPlugins)
+		{
+			if(plugin.Info.dwPluginId1 == PLUGMAGIC('V', 's', 't', 'P'))
+			{
+				m_playBehaviour.set(kLegacyPPQpos);
+				break;
+			}
+		}
+	}
+
+	if(m_dwLastSavedWithVersion < MPT_V("1.32.00.38") && hasAnyPlugins)
+	{
+		for(const auto &plugin : m_MixPlugins)
+		{
+			if(plugin.Info.dwPluginId1 == PLUGMAGIC('V', 's', 't', 'P'))
+			{
+				m_playBehaviour.set(kLegacyPluginNNABehaviour);
+				break;
+			}
+		}
+	}
+#endif
 }
 
 OPENMPT_NAMESPACE_END
